""" jouissance.tf.loss """

import tensorflow as tf

from jouissance.util import read_conf


def genopt(con=None):
    """ generator optimizer """
    con = read_conf() if con is None else con
    _opt = tf.keras.optimizers.Adam(
        learning_rate=con["genopt_learning_rate"],
        beta_1=con["genopt_beta1"],
        beta_2=con["genopt_beta2"]
    )
    return tf.keras.mixed_precision.LossScaleOptimizer(
        _opt) if con["use_amp"] else _opt


def disopt(con=None):
    """ discriminator optimizer """
    con = read_conf() if con is None else con
    _opt = tf.keras.optimizers.Adam(
        learning_rate=con["disopt_learning_rate"],
        beta_1=con["disopt_beta1"],
        beta_2=con["disopt_beta2"]
    )
    return tf.keras.mixed_precision.LossScaleOptimizer(
        _opt) if con["use_amp"] else _opt


def genloss(fake_log, real_img, fake_img):
    """ loss function for generator """
    return -tf.reduce_mean(fake_log) + tf.reduce_mean(
        tf.keras.losses.mse(real_img, fake_img))


def disloss(real_log, fake_log, real_img, fake_img):
    """ loss function for discriminator """
    real_loss = tf.reduce_mean(real_log)
    fake_loss = tf.reduce_mean(fake_log)
    return fake_loss - real_loss - tf.reduce_mean(
        tf.keras.losses.mse(real_img, fake_img))



""" test: jouissance.util.conf """

import pytest

from jouissance.util.conf import read_conf, make_hash, edit_conf


def test_read_conf():
    """ test: read the yaml configs """

    assert read_conf()["default"]

    with pytest.raises(FileNotFoundError):
        read_conf(conf_path="/tmp/some_configs.yaml")


def test_make_hash():
    """ test: make a hash from the dict for caching """
    assert make_hash("test") != "something"


def test_edit_conf():
    """ test: edit the configs and export them """
    con = read_conf()
    fn_con = edit_conf(con)
    assert con == read_conf(fn_con)
    new_con = con.copy()
    new_con["something_new"] = "hello"
    fn_new_con = edit_conf(new_con)
    assert read_conf(fn_new_con)["something_new"] == "hello"



""" test: jouissance.util.file """

from jouissance.util.conf import read_conf
from jouissance.util.file import fsspec_fs, make_name, parse_days, make_glob
from jouissance.util.file import get_scenes_conditions_h5py, save_conditions
from jouissance.util.file import save_scenes, get_scenes_conditions_netcdf4


def test_fsspec_fs():
    """ test: get the fsspec filesystem """

    assert fsspec_fs(provider="gs").protocol[0] in ["gs", "gcs"]
    assert fsspec_fs(provider="s3").protocol[0] == "s3"
    assert fsspec_fs(provider="abfs").protocol == "abfs"
    assert fsspec_fs(provider="file").protocol == "file"


def test_make_name():
    """ test: get the provider from the path """

    assert make_name("gs", "goes-17") == "gcp-public-data-goes-17"
    assert make_name("gcs", "goes-17") == "gcp-public-data-goes-17"
    assert make_name("s3", "goes-17") == "noaa-goes17"
    assert make_name("abfs", "goes-17") == "noaa-goes17"
    assert make_name("abfs", "GOES-17") == "noaa-goes17"


def test_parse_days():
    """ test: return a list of applicable days """

    assert parse_days("111") == ["111"]
    assert parse_days("11*")[0] == "110"
    assert parse_days("11?")[-1] == "119"
    assert len(parse_days("11*")) == 10
    assert len(parse_days("***")) == 365


def test_make_glob():
    """ test: make a pattern for glob and return glob """

    _az = fsspec_fs(provider="abfs")
    _gs = fsspec_fs(provider="gs")
    _s3 = fsspec_fs(provider="s3")
    azdn = "noaa-goes17"
    gsdn = "gcp-public-data-goes-17"
    s3dn = "noaa-goes17"

    assert len(make_glob(_az, azdn)) == 12
    assert len(make_glob(_gs, gsdn)) == 12
    assert len(make_glob(_s3, s3dn)) == 12


def test_get_scenes_conditions_h5py():
    """ test h5py/netcdf4 scenes/conditions """

    save_conditions()
    con = read_conf()
    _fs = fsspec_fs(con=con)
    _dn = make_name("gs", "goes-17", con=con)
    _gb = make_glob(_fs, _dn, con=con)

    for _tool in ["h5py", "netcdf4"]:

        if _tool == "netcdf4":
            save_scenes(_gb, _fs, con=con)
            _goal = _gb[0].split("/")[-1]
            get_scenes_conditions = get_scenes_conditions_netcdf4
        else:
            _goal = _gb[0]
            get_scenes_conditions = get_scenes_conditions_h5py

        scenes, conditions = get_scenes_conditions(_goal)

        assert scenes.shape == (
            int(con["scene_shape1"]),
            int(con["scene_shape2"]),
            len(con["channels"])
        )

        assert conditions.shape == (
            int(con["reanalysis_shape1"]),
            int(con["reanalysis_shape2"]),
            len(con["reanalysis_levels"]) * len(con["reanalysis_products"])
        )



""" test: jouissance.pt.data """

import os

if os.environ.get("PDGM_BACK", "tf") in ["pt", "cf"]:
    from jouissance.pt.data import make_data
    from jouissance.util.file import save_conditions


def test_make_data():
    """ test: make data """

    if os.environ.get("PDGM_BACK", "tf") not in ["pt", "cf"]:
        return

    save_conditions()

    tds = make_data()

    for val in tds:
        assert val[0].shape == (2, 2, 32, 64)
        assert val[1].shape == (2, 27, 32, 64)



""" jouissance.tf.data """

import tensorflow as tf

from jouissance.util import read_conf, make_hash
from jouissance.util import get_scenes_conditions_h5py
from jouissance.util import make_glob


class MyData(tf.data.Dataset):  # pylint: disable=abstract-method
    """ https://www.tensorflow.org/guide/data_performance """

    def _generator(self):
        yield get_scenes_conditions_h5py(self)

    def __new__(cls, file_paths=None):
        if file_paths is None:
            raise ValueError("file_paths is required")
        con = read_conf()
        scenes_shape = (
            int(con["scene_shape1"]), int(con["scene_shape2"]),
            len(con["channels"])
        )
        conditions_shape = (
            int(con["reanalysis_shape1"]), int(con["reanalysis_shape2"]),
            len(con["reanalysis_products"])*len(con["reanalysis_levels"])
        )
        return tf.data.Dataset.from_generator(
            cls._generator,
            output_signature=(
                tf.TensorSpec(
                    shape=tf.TensorShape(scenes_shape),
                    dtype=tf.float32),
                tf.TensorSpec(
                    shape=tf.TensorShape(conditions_shape),
                    dtype=tf.float32),
            ),
            args=(file_paths,),
        )


def make_data():
    """ https://www.tensorflow.org/guide/data_performance """
    configs = read_conf()

    files = make_glob(con=configs)

    nsamples = len(files)
    tds = tf.data.Dataset.from_tensor_slices(files)
    # pylint: disable=abstract-class-instantiated
    tds = tds.interleave(
        lambda file: MyData(file),  # pylint: disable=unnecessary-lambda
        num_parallel_calls=tf.data.experimental.AUTOTUNE,
        deterministic=False,
    )

    sshape = (
        int(configs["scene_reshape1"]),
        int(configs["scene_reshape2"])
    )
    cshape = (
        int(configs["reanalysis_reshape1"]),
        int(configs["reanalysis_reshape2"])
    )
    tds = tds.map(
        lambda _scene, _condition: (
            tf.image.resize(_scene, sshape),
            tf.image.resize(_condition, cshape)
        ), num_parallel_calls=tf.data.experimental.AUTOTUNE
    )

    tds = tds.map(lambda _scene, _cond: (
        (_scene - tf.math.reduce_mean(_scene, axis=[1, 2], keepdims=True)) /
        tf.math.reduce_std(_scene, axis=[1, 2], keepdims=True),
        (_cond - tf.math.reduce_min(_cond, axis=[1, 2], keepdims=True)) /
        (
            tf.math.reduce_max(_cond, axis=[1, 2], keepdims=True) -
            tf.math.reduce_min(_cond, axis=[1, 2], keepdims=True)
        )
    ), num_parallel_calls=tf.data.experimental.AUTOTUNE)

    tds = tds.batch(
        batch_size=int(configs["batch_size"]), drop_remainder=True
    )

    tds = tds.prefetch(tf.data.AUTOTUNE)

    train_take = int(
        float(configs["train_split"]) * nsamples // int(configs["batch_size"])
    )

    cache_file = configs["cache_prefix"] + make_hash(configs)

    if configs["if_shuffle"]:
        training_ds = tds.take(train_take).cache(cache_file).shuffle(
            int(configs["shuffle_size"]),
            reshuffle_each_iteration=configs["reshuffle_iter"],
        )
        validating_ds = tds.skip(train_take).cache(cache_file).shuffle(
            int(configs["shuffle_size"]),
            reshuffle_each_iteration=configs["reshuffle_iter"],
        )
    else:
        training_ds = tds.take(train_take).cache(cache_file)
        validating_ds = tds.skip(train_take).cache(cache_file)

    return training_ds, validating_ds



""" jouissance.pt.data """

import os
import glob

import torch
import torchvision

from jouissance.util import read_conf, make_hash
from jouissance.util import get_scenes_conditions_h5py
from jouissance.util import make_glob


class MyData(torch.utils.data.Dataset):
    """ custom dataset """

    def __init__(self, files, _hash, configs):
        self.files = files
        self.sfiles = [file.split("/")[-1] for file in files]
        self._hash = _hash
        self.con = configs

    def __len__(self):
        return len(self.files)

    def __getitem__(self, index):
        if (
            os.path.exists(f"{self._hash}/scenes{self.sfiles[index]}") and
            os.path.exists(f"{self._hash}/conditions{self.sfiles[index]}")
        ):
            return (
                torch.load(
                    f"{self._hash}/scenes{self.sfiles[index]}"),
                torch.load(
                    f"{self._hash}/conditions{self.sfiles[index]}")
            )

        scenes, conditions = get_scenes_conditions_h5py(
            self.files[index], bck="pt")
        scenes, conditions = torch.Tensor(scenes), torch.Tensor(conditions)

        scenes = torchvision.transforms.Resize((
            int(self.con["scene_reshape1"]),
            int(self.con["scene_reshape2"])
        ))(scenes)

        conditions = torchvision.transforms.Resize((
            int(self.con["reanalysis_reshape1"]),
            int(self.con["reanalysis_reshape2"])
        ))(conditions)

        os.makedirs(f"{self._hash}", exist_ok=True)

        torch.save(
            scenes, f"{self._hash}/scenes{self.sfiles[index]}")
        torch.save(
            conditions, f"{self._hash}/conditions{self.sfiles[index]}")

        return scenes, conditions


def make_data():
    """ make a dataset """

    configs = read_conf()
    cache_fold = f"{configs['cache_prefix']}{make_hash(configs)}pt"

    if configs["use_cache_only"]:
        files = glob.glob(f"{cache_fold}/scenes*")
        files = [file.replace("scenes", "") for file in files]
    else:
        files = make_glob(con=configs)

    return torch.utils.data.DataLoader(
        MyData(files, cache_fold, configs),
        batch_size=int(configs["batch_size"]),
        shuffle=True,
        num_workers=0,
    )



""" jouissance.tf.arch """

import numpy as np
import tensorflow as tf


def set_initializer():
    """ set the initializer """
    return tf.random_normal_initializer(0.0, 0.02)


def downsample_block(
    filters, size, strides=2, drop_rate=0.0, apply_batchnorm=True,
):
    """ a downsampling block. """
    initializer = set_initializer()

    result = tf.keras.Sequential()

    result.add(tf.keras.layers.Conv2D(
        filters, size, strides=strides, padding="same",
        kernel_initializer=initializer, use_bias=False,
        data_format="channels_last"
    ))

    if apply_batchnorm:
        result.add(tf.keras.layers.BatchNormalization())

    if drop_rate > 0:
        result.add(tf.keras.layers.Dropout(drop_rate))

    result.add(tf.keras.layers.PReLU())

    return result


def upsample_block(
    filters, size, strides=2, drop_rate=0.0, act="normal",
):
    """ an upsampling block. """
    initializer = set_initializer()

    transpose = False

    result = tf.keras.Sequential()

    if transpose:
        result.add(tf.keras.layers.Conv2DTranspose(
            filters, size, strides=strides, padding="same",
            kernel_initializer=initializer, use_bias=False,
            data_format="channels_last"
        ))

    if not transpose:
        result.add(tf.keras.layers.UpSampling2D(
            size=strides, interpolation="bilinear",
        ))

    result.add(tf.keras.layers.BatchNormalization())

    if drop_rate > 0:
        result.add(tf.keras.layers.Dropout(drop_rate))

    if act == "exit":
        result.add(tf.keras.layers.Activation("tanh"))
    else:
        result.add(tf.keras.layers.PReLU())

    return result


def dense_disc(
    conditions_shape=(128, 256, 3),
    scenes_shape=(128, 256, 1),
    name=None,
):
    """ a dense discriminator """

    input_conditions = tf.keras.layers.Input(shape=conditions_shape)
    if scenes_shape[0] != conditions_shape[0] \
            or scenes_shape[1] != conditions_shape[1]:
        holding = upsample_block(
            filters=3,
            size=5,
            strides=(
                scenes_shape[0]//conditions_shape[0],
                scenes_shape[1]//conditions_shape[1]
            )
        )(input_conditions)
    else:
        holding = input_conditions
    input_scenes = tf.keras.layers.Input(shape=scenes_shape)
    inputs = tf.keras.layers.Concatenate()([input_scenes, holding])
    output = tf.keras.layers.BatchNormalization()(inputs)
    steps = int(np.log2(np.min([scenes_shape[0], scenes_shape[1]])))
    for _ in range(steps):
        output = downsample_block(
            filters=64, size=3, drop_rate=0.2,
            apply_batchnorm=False)(output)
        residual = output
        output = downsample_block(
            filters=64, size=3, strides=1,
            apply_batchnorm=False)(output)
        output = tf.keras.layers.add([residual, output])
        output = tf.keras.layers.BatchNormalization()(output)

    output = tf.keras.layers.Flatten()(output)
    output = tf.keras.layers.Dropout(.2)(output)

    output = tf.keras.layers.Dense(
        1,
        activation="tanh", dtype='float32'
    )(output)

    return tf.keras.Model(
        inputs=[input_conditions, input_scenes],
        outputs=[output], name=name)


def unet_model(
    conditions_shape=(128, 256, 3),
    scenes_shape=(128, 256, 1),
    use_latent=True,
    name=None,
):
    """ a unet-like generation model. """
    input_conditions = tf.keras.layers.Input(shape=conditions_shape)
    if use_latent:
        input_noise = tf.keras.layers.Input(shape=conditions_shape)
        inputs = tf.keras.layers.Concatenate()([input_conditions, input_noise])
    else:
        inputs = input_conditions
    output = inputs

    skips = []
    steps = int(np.log2(np.min([conditions_shape[0], conditions_shape[1]])))
    for _ in range(steps):
        output = downsample_block(
            filters=64, size=3, drop_rate=0.2)(output)
        residual = output
        output = downsample_block(
            filters=64, size=3, strides=1)(output)
        output = tf.keras.layers.add([residual, output])
        skips.append(output)

    skips = reversed(skips[:-1])

    stepsagain = int(np.log2(np.min([scenes_shape[0], scenes_shape[1]])))
    for _, skip in zip(range(steps-1), skips):
        output = upsample_block(
            filters=64, size=3, drop_rate=0.2,)(output)
        residual = output
        output = upsample_block(
            filters=64, size=3, strides=1,)(output)
        output = tf.keras.layers.add([residual, output])
        output = tf.keras.layers.Concatenate()([output, skip])

    passcounter = 1
    for _ in range(steps, stepsagain):
        if passcounter == 1:
            output = upsample_block(
                filters=32, size=5, strides=4,)(output)
            residual = output
            output = upsample_block(
                filters=32, size=5, strides=1,)(output)
            output = tf.keras.layers.add([residual, output])
        passcounter = passcounter*-1

    if abs(stepsagain-steps) % 2 == 0:
        output = upsample_block(filters=32, size=3, strides=2,)(output)

    output = tf.keras.layers.Conv2D(
        scenes_shape[-1],
        3,
        strides=1,
        padding="same",
        activation="sigmoid",
        use_bias=True,
        data_format="channels_last"
    )(output)

    output = tf.keras.layers.Reshape(scenes_shape, dtype='float32')(output)
    if use_latent:
        return tf.keras.Model(
            inputs=[input_conditions, input_noise],
            outputs=[output], name=name)
    return tf.keras.Model(inputs=input_conditions, outputs=[output], name=name)



""" test: jouissance.pt.arch """

import os

if os.environ.get("PDGM_BACK", "tf") in ["pt", "cf"]:
    import torch
    from jouissance.pt.arch import DownSample, UpSample, DenseDisc, UNetModel


def test_downsample_block():
    """ test: a downsampling block """
    if os.environ.get("PDGM_BACK", "tf") == "tf":
        return
    result = DownSample(
        _in=128, filters=64, size=3, stride=2, drop_rate=0.0, apply_batchnorm=True,
    )
    assert result(
        torch.rand((2, 128, 32, 64))  # pylint: disable=no-member
    ).shape == (2, 64, 16, 32)


def test_upsample_block():
    """ test: an upsampling block """
    if os.environ.get("PDGM_BACK", "tf") == "tf":
        return
    result = UpSample(
        _in=1, stride=2, drop_rate=0.0, act="normal",
    )
    assert result(
        torch.rand((1, 1, 1, 1))  # pylint: disable=no-member
    ).shape == (1, 1, 2, 2)


def test_dense_disc():
    """ test: dense disc """
    # pylint: disable=no-member
    if os.environ.get("PDGM_BACK", "tf") == "tf":
        return
    _con1 = torch.rand((5, 5, 32, 64))
    _sce = torch.rand((5, 1, 64, 128))
    disc1 = DenseDisc(
        scenes_shape=list(_sce.shape[1:]),
        conditions_shape=list(_con1.shape[1:]),
    )
    assert tuple(disc1(_con1, _sce).shape) == (5, 1)
    _con2 = torch.rand((5, 5, 64, 128))
    _sce = torch.rand((5, 1, 64, 128))
    disc2 = DenseDisc(
        scenes_shape=list(_sce.shape[1:]),
        conditions_shape=list(_con2.shape[1:]),
    )
    assert tuple(disc2(_con2, _sce).shape) == (5, 1)


def test_unet_model():
    """ test: unet generator """
    # pylint: disable=no-member
    if os.environ.get("PDGM_BACK", "tf") == "tf":
        return
    scenes = torch.rand((5, 1, 32, 64))
    conditions = torch.rand((5, 5, 32, 64))
    unet_no_latent = UNetModel(
        scenes_shape=scenes.shape[1:],
        conditions_shape=conditions.shape[1:],
        use_latent=False,
    )
    unet_latent = UNetModel(
        scenes_shape=scenes.shape[1:],
        conditions_shape=conditions.shape[1:],
        use_latent=True,
    )
    assert unet_no_latent(conditions).shape == scenes.shape
    assert unet_latent(conditions, conditions).shape == scenes.shape


def test_unet_shapes():
    """ test: unet model """
    # pylint: disable=no-member
    if os.environ.get("PDGM_BACK", "tf") == "tf":
        return
    conditions = [
        (9, 2**idx, 2**(jdx+1)) for idx, jdx in zip(range(2, 7), range(2, 7))
    ]
    scenes = [
        (3, 2**idx, 2**(jdx+1)) for idx, jdx in zip(range(2, 7), range(2, 7))
    ]

    for idx, con in enumerate(conditions):
        for sce in scenes[idx:]:
            csample = torch.rand((5, 9, con[1], con[2]))
            ssample = torch.rand((5, 3, sce[1], sce[2]))
            unet_latent = UNetModel(
                scenes_shape=sce,
                conditions_shape=con,
                use_latent=True
            )
            unet_no_latent = UNetModel(
                scenes_shape=sce,
                conditions_shape=con,
                use_latent=False
            )
            assert unet_no_latent(csample).shape == ssample.shape
            assert unet_latent(csample, csample).shape == ssample.shape



""" jouissance.util.file """

import os
import urllib
import urllib.request
import itertools

import cftime
import fsspec
import netCDF4
import h5py
import numpy as np

from jouissance.util.conf import read_conf


def fsspec_fs(provider=None, con=None):
    """ get the fsspec filesystem """

    con = read_conf() if con is None else con
    provider = con["provider"] if provider is None else provider

    if provider == "abfs":
        return fsspec.filesystem(
            provider, account_name=con["abfs_account_name"])

    if con["provider"] == "file":
        return fsspec.filesystem(
            provider)

    return fsspec.filesystem(
        provider, token="anon", anon=True)


def make_name(_pr=None, _dn=None, con=None):
    """ get the provider from the path """

    con = read_conf() if con is None else con
    _pr = con["provider"] if _pr is None else _pr
    _dn = "goes-17" if _dn is None else _dn

    if _pr.lower() not in ["gs", "gcs", "s3", "abfs", "file"]:
        raise ValueError(f"{_pr} must be gs, s3, abfs, or file")

    if _pr.lower() == "file":
        return con["data_folder"]

    if not _dn.lower().startswith("goes"):
        raise ValueError("only goes-?? data is supported")

    return f"gcp-public-data-{_dn.lower()}" if _pr.lower().startswith("g") \
        else "noaa-" + _dn.lower().replace("-", "")


def parse_days(_dys):
    """ return a list of applicable days """

    _dys = _dys*3 if len(_dys) < 3 else _dys
    pos0, pos1, pos2 = _dys[0], _dys[1], _dys[2]
    perms = list(itertools.product('0123456789', repeat=3))
    return [
        f"{perm[0]}{perm[1]}{perm[2]}"
        for perm in perms if
        (perm[0] == str(pos0) or pos0 in ["?", "*"]) and
        (perm[1] == str(pos1) or pos1 in ["?", "*"]) and
        (perm[2] == str(pos2) or pos2 in ["?", "*"]) and
        int(f"{perm[0]}{perm[1]}{perm[2]}") > 0 and
        int(f"{perm[0]}{perm[1]}{perm[2]}") < 366
    ]


def make_glob(_fs=None, _dn=None, con=None):
    """ make a glob """

    con = read_conf() if con is None else con
    _dn = make_name(con=con) if _dn is None else _dn
    _fs = fsspec_fs(con=con) if _fs is None else _fs

    _px = con["products"][0]
    _gb = []

    if len(con["days"]) < 3 or "?" in con["days"] or "*" in con["days"]:
        _dys = parse_days(con["days"][0])
    else:
        _dys = con["days"]
    for _yr in con["years"]:
        for _hr in con["hours"]:
            for _dy in _dys:
                _gb += _fs.glob(
                    f"{_dn}/*{_px}*{_yr}{_dy}{_hr}*.nc"
                ) if con["provider"] == "local" else _fs.glob(
                    f"{_dn}/{_px}/{_yr}/{_dy}/{_hr}/*.nc"
                )

    return _gb


def get_scenes_conditions_netcdf4(file, bck="tf", con=None):
    """ return the scenes and conditions """
    # pylint: disable=no-member

    con = read_conf() if con is None else con

    file = file if isinstance(file, str) else file.decode("utf-8")

    scenes, conditions = [], []

    _ncs = netCDF4.Dataset(f"{con['scene_folder']}/{file}")

    scenes.extend(
        np.expand_dims(_ncs[f"CMI_C{_ch}"][:].data, axis=-1)
        for _ch in con["channels"]
    )

    _nctime = _ncs["t"][:].data.copy()
    _ncunit = _ncs["t"].units
    time = cftime.num2date(_nctime, _ncunit)
    _ci = (time.dayofyr - 1)*4 + (time.hour//6)
    _ncs.close()

    for _pd in con["reanalysis_products"]:
        _ncc = netCDF4.Dataset(
            f"{con['cond_folder']}/{_pd}.{time.year}.nc"
        )
        conditions.append(
            _ncc[_pd][_ci, con["reanalysis_levels"], :, :].data.copy())
        _ncc.close()

    if bck == "pt":
        return (
            np.moveaxis(np.concatenate(scenes, axis=-1), -1, 0),
            np.concatenate(conditions)
        )

    return (
        np.concatenate(scenes, axis=-1),
        np.moveaxis(np.concatenate(conditions), 0, -1)
    )


def get_scenes_conditions_h5py(file, bck="tf", con=None):
    """ return the scenes and conditions """
    # pylint: disable=no-member

    con = read_conf() if con is None else con
    _fs = fsspec_fs()

    file = file if isinstance(file, str) else file.decode("utf-8")

    scenes, conditions = [], []

    _ncs = h5py.File(_fs.open(file, "rb"), "r")

    scenes.extend(
        np.expand_dims(
            _ncs[f"CMI_C{_ch}"][:] *
            _ncs[f"CMI_C{_ch}"].attrs["scale_factor"] +
            _ncs[f"CMI_C{_ch}"].attrs["add_offset"],
            axis=-1
        ) for _ch in con["channels"]
    )

    _nctime = _ncs["t"][()]
    _ncunit = _ncs["t"].attrs["units"].decode("utf-8")
    time = cftime.num2date(_nctime, _ncunit)
    _ci = (time.dayofyr - 1)*4 + (time.hour//6)
    _ncs.close()

    for _pd in con["reanalysis_products"]:
        _ncc = h5py.File(
            f"{con['cond_folder']}/{_pd}.{time.year}.nc"
        )
        conditions.append(
            _ncc[_pd][_ci, con["reanalysis_levels"], :, :].copy())
        _ncc.close()

    if bck == "pt":
        return (
            np.moveaxis(np.concatenate(scenes, axis=-1), -1, 0),
            np.concatenate(conditions)
        )

    return (
        np.concatenate(scenes, axis=-1),
        np.moveaxis(np.concatenate(conditions), 0, -1)
    )


def save_conditions(con=None):
    """ save the conditions """

    con = read_conf() if con is None else con
    _pxs = con["reanalysis_products"]
    _yrs = con["years"]
    _vr = con["reanalysis_version"]
    _fol = con["cond_folder"]
    version = "" if _vr == "1" else f"{_vr}"
    _fns = [f"{_px}.{_yr}.nc" for _px in _pxs for _yr in _yrs]
    base_url = "https://downloads.psl.noaa.gov/Datasets/ncep.reanalysis"
    for _fn in _fns:
        if not os.path.exists(f"{_fol}/{_fn}"):
            os.makedirs(_fol, exist_ok=True)
            urllib.request.urlretrieve(
                f"{base_url}{version}/pressure/{_fn}",
                f"{_fol}/{_fn}")


def save_scenes(_glb=None, _fs=None, con=None):
    """ save scenes via glob """

    con = read_conf() if con is None else con
    _fs = fsspec_fs(con=con) if _fs is None else _fs
    _glb = make_glob(_fs, con=con) if _glb is None else _glb

    _fol = con["scene_folder"]

    for _fn in _glb:
        _fnx = _fn.split("/")[-1]
        if not os.path.exists(f"{_fol}/{_fnx}"):
            os.makedirs(_fol, exist_ok=True)
            _fs.get(_fn, f"{_fol}/{_fnx}")



""" jouissance.util.conf: configs utils """

import hashlib
import json
import os

import yaml


def read_conf(conf_path=None):
    """ read the yaml configs """

    if conf_path is not None:
        return yaml.safe_load(open(conf_path, "r", encoding="utf8"))

    if os.environ.get("PDGM_CONFIGS_PATH") is not None:
        conf_path = os.environ.get("PDGM_CONFIGS_PATH")
        return yaml.safe_load(open(conf_path, "r", encoding="utf8"))

    con_dir = os.environ.get("PDGM_CONFIGS_DIR", os.getcwd())

    configs_path = f"{con_dir}/configs.yaml" if \
        os.path.exists(f"{con_dir}/configs.yaml") else \
        f"{os.path.dirname(__file__)}/_configs.yaml"

    return yaml.safe_load(open(configs_path, "r", encoding="utf8"))


def make_hash(con=None):
    """ make a hash from the dict for caching """

    _hash = hashlib.md5()
    dumped = json.dumps(con, sort_keys=True).encode()
    _hash.update(dumped)

    return f".cache{_hash.hexdigest()}"


def edit_conf(con):
    """ edit the configs and export them """

    filename = make_hash(con)

    with open(f"{filename}.yaml", "w", encoding="utf8") as _fo:
        yaml.dump(con, _fo)

    return f"{filename}.yaml"



""" jouissance.pt.arch """

import numpy as np
import torch


class DownSample(torch.nn.Module):
    """ a downsampling block """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        _in, filters, size,
        stride=2, apply_batchnorm=True, drop_rate=0.0
    ):
        """ definitions """

        super().__init__()

        self.main = torch.nn.Sequential()

        self.main.add_module(
            "conv2d",
            torch.nn.Conv2d(
                _in, filters, size, stride=stride, padding=(1, 1), bias=False
            )
        )

        if apply_batchnorm:
            self.main.add_module(
                "batchnorm2d",
                torch.nn.BatchNorm2d(filters)
            )

        if drop_rate > 0:
            self.main.add_module(
                "drop",
                torch.nn.Dropout2d(drop_rate)
            )

        self.main.add_module(
            "act",
            torch.nn.PReLU()
        )

    def forward(self, _input):
        """ move """

        return self.main(_input)


class UpSample(torch.nn.Module):
    """ an upsampling block """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        _in,
        stride=2, drop_rate=0.0, act="normal"
    ):
        """ definitions """

        super().__init__()

        self.main = torch.nn.Sequential()

        self.main.add_module(
            "upsample",
            torch.nn.Upsample(
                scale_factor=stride, mode="bilinear",
            )
        )

        self.main.add_module(
            "batchnorm2d",
            torch.nn.BatchNorm2d(_in)
        )

        if drop_rate > 0:
            self.main.add_module(
                "dropout",
                torch.nn.Dropout2d(drop_rate)
            )

        if act == "exit":
            self.main.add_module(
                "tanh",
                torch.nn.Tanh()
            )
        else:
            self.main.add_module(
                "prelu",
                torch.nn.PReLU()
            )

    def forward(self, _input):
        """ move """

        return self.main(_input)


class DenseDisc(torch.nn.Module):
    """ dense disc """

    def __init__(
        self,
        conditions_shape=(3, 128, 256),
        scenes_shape=(1, 128, 256),
    ):
        """ defs """
        super().__init__()
        self.conditions_shape = conditions_shape
        self.scenes_shape = scenes_shape

    def forward(self, _con, _sce):
        """ move """
        # pylint: disable=no-member
        hold = _con
        if self.conditions_shape[1] != self.scenes_shape[1] \
                or self.conditions_shape[2] != self.scenes_shape[2]:
            hold = UpSample(
                self.conditions_shape[0],
                stride=(
                    self.scenes_shape[1]//self.conditions_shape[1],
                    self.scenes_shape[2]//self.conditions_shape[2]
                )
            )(_con)

        _input = torch.cat([_sce, hold], axis=1)
        _out = torch.nn.BatchNorm2d(
            self.conditions_shape[0]+self.scenes_shape[0])(_input)
        steps = int(np.log2(np.min(
            [self.scenes_shape[1], self.scenes_shape[2]])))
        feats = self.conditions_shape[0]+self.scenes_shape[0]
        for _ in range(steps):
            _out = DownSample(
                feats,
                filters=64, size=3, drop_rate=0.2, stride=2,
                apply_batchnorm=False,
            )(_out)
            feats = 64
            _res = _out
            _out = DownSample(
                feats,
                filters=64, size=3, drop_rate=0.2, stride=1,
                apply_batchnorm=False,
            )(_out)
            _out = torch.add(_res, _out)
            _out = torch.nn.BatchNorm2d(feats)(_out)

        _out = torch.nn.Flatten()(_out)
        _out = torch.nn.Dropout(0.2)(_out)

        _out = torch.nn.LazyLinear(1, dtype=torch.float32)(_out)

        _out = torch.nn.Tanh()(_out)

        return _out


class UNetModel(torch.nn.Module):
    """ a unet-like generation model """

    def __init__(
        self,
        conditions_shape=(3, 128, 256),
        scenes_shape=(1, 128, 256),
        use_latent=True,
    ):
        """ defs """
        super().__init__()
        self.conditions_shape = conditions_shape
        self.scenes_shape = scenes_shape
        self.use_latent = use_latent

    def forward(self, _con, _lat=None):
        """ move """
        # pylint: disable=no-member
        _input = _con
        if self.use_latent:
            if _lat is None:
                raise ValueError("latent var is requires")
            _input = torch.cat([_con, _lat], axis=1)
        skips = []
        steps = int(np.log2(np.min(
            [self.conditions_shape[-2], self.conditions_shape[-1]]
        )))
        feats = self.conditions_shape[0] + \
            self.conditions_shape[0] * (self.use_latent)
        _out = _input
        for _ in range(steps):
            _out = DownSample(
                feats,
                filters=64, size=3, drop_rate=0.2, stride=2)(_out)
            feats = 64
            _res = _out
            _out = DownSample(
                feats,
                filters=64, size=3, stride=1)(_out)
            _out = torch.add(_res, _out)
            skips.append(_out)

        skips = reversed(skips[:-1])

        stepsagain = int(np.log2(np.min(
            [self.scenes_shape[-2], self.scenes_shape[-1]]
        )))
        for _, skip in zip(range(steps-1), skips):
            _out = UpSample(
                feats,
                drop_rate=0.2, stride=2)(_out)
            _res = _out
            _out = UpSample(
                feats,
                stride=1)(_out)
            _out = torch.add(_res, _out)
            _out = torch.cat([_out, skip], axis=1)
            feats += skip.shape[1]

        passcounter = 1
        for _ in range(steps, stepsagain):
            if passcounter == 1:
                _out = UpSample(
                    feats, stride=4)(_out)
                _res = _out
                _out = UpSample(
                    feats, stride=1)(_out)
                _out = torch.add(_res, _out)
            passcounter = passcounter*-1

        if abs(stepsagain-steps) % 2 == 0:
            _out = UpSample(
                feats, stride=2)(_out)

        _out = torch.nn.Conv2d(
            feats, self.scenes_shape[0], 3, stride=1, padding=(1, 1), bias=True
        )(_out)
        _out = torch.nn.Sigmoid()(_out)

        return _out



""" test: jouissance.tf.data """

import os

from jouissance.util import read_conf, edit_conf, save_conditions

if os.environ.get("PDGM_BACK", "tf") in ["tf", "cf", "jx"]:
    from jouissance.tf import MyData, make_data


def test_my_data():
    """ test: MyData class """
    if os.environ.get("PDGM_BACK", "tf") == "pt":
        return
    _path = "gcp-public-data-goes-17/ABI-L2-CMIPC/" + \
        "2020/001/00/OR_ABI-L2-CMIPC-M6C01_G17_" + \
        "s20200010046212_e20200010048585_c20200010049042.nc"
    # pylint: disable=abstract-class-instantiated
    assert MyData(_path).element_spec[0].shape == (1500, 2500, 2)
    assert MyData(_path).element_spec[1].shape == (73, 144, 27)
    con = read_conf()
    new_con = con.copy()
    new_con["reanalysis_presave"] = True
    _fn_con = edit_conf(new_con)
    os.environ["PDGM_CONFIGS_PATH"] = _fn_con
    assert MyData(_path).element_spec[0].shape == (1500, 2500, 2)
    assert MyData(_path).element_spec[1].shape == (73, 144, 27)
    new_con["reanalysis_presave"] = False
    _fn_con = edit_conf(new_con)
    os.environ["PDGM_CONFIGS_PATH"] = _fn_con


def test_make_data():
    """ test: make_data function """
    if os.environ.get("PDGM_BACK", "tf") == "pt":
        return
    assert make_data()[0].element_spec[0].shape == (2, 32, 64, 2)
    assert make_data()[0].element_spec[1].shape == (2, 32, 64, 27)
    con = read_conf()
    new_con = con.copy()
    new_con["reanalysis_presave"] = True
    _fn_con = edit_conf(new_con)
    os.environ["PDGM_CONFIGS_PATH"] = _fn_con
    assert make_data()[0].element_spec[0].shape == (2, 32, 64, 2)
    assert make_data()[0].element_spec[1].shape == (2, 32, 64, 27)
    new_con["reanalysis_presave"] = False
    _fn_con = edit_conf(new_con)
    os.environ["PDGM_CONFIGS_PATH"] = _fn_con


def test_make_data_loop():
    """ test: make_data function with loops """
    if os.environ.get("PDGM_BACK", "tf") == "pt":
        return
    save_conditions()
    tds, _ = make_data()
    for idx, jdx in tds:
        assert idx.shape == (2, 32, 64, 2)
        assert jdx.shape == (2, 32, 64, 27)
        break

    con = read_conf()
    new_con = con.copy()
    new_con["reanalysis_presave"] = True
    _fn_con = edit_conf(new_con)
    os.environ["PDGM_CONFIGS_PATH"] = _fn_con

    tds2, _ = make_data()
    for idx, jdx in tds2:
        assert idx.shape == (2, 32, 64, 2)
        assert jdx.shape == (2, 32, 64, 27)
        break

    new_con["reanalysis_presave"] = False
    _fn_con = edit_conf(new_con)
    os.environ["PDGM_CONFIGS_PATH"] = _fn_con



""" test: jouissance.tf.arch """

import os

if os.environ.get("PDGM_BACK", "tf") in ["tf", "cf", "jx"]:
    import tensorflow as tf

    from jouissance.tf import set_initializer
    from jouissance.tf import upsample_block, downsample_block
    from jouissance.tf import dense_disc, unet_model


def test_set_initializer():
    """ test: set the initializer """
    if os.environ.get("PDGM_BACK", "tf") == "pt":
        return
    assert set_initializer().mean == 0.
    assert set_initializer().stddev == 0.02


def test_upsample_block():
    """ test: an upsampling block """
    if os.environ.get("PDGM_BACK", "tf") == "pt":
        return
    result = upsample_block(
        filters=64, size=3, strides=2, drop_rate=0.0, act="normal",
    )
    assert result(
        tf.random.normal(shape=(1, 1, 1, 1))
    ).shape == (1, 2, 2, 1)


def test_downsample_block():
    """ test: a downsampling block """
    if os.environ.get("PDGM_BACK", "tf") == "pt":
        return
    result = downsample_block(
        filters=64, size=3, strides=2, drop_rate=0.0, apply_batchnorm=True,
    )
    assert result(
        tf.random.normal(shape=(1, 32, 64, 128))
    ).shape == (1, 16, 32, 64)


def test_dense_disc():
    """ test: dense disc """
    if os.environ.get("PDGM_BACK", "tf") == "pt":
        return
    scenes = tf.random.normal(shape=(5, 32, 64, 1))
    conditions = tf.random.normal(shape=(5, 32, 64, 5))
    disc = dense_disc(
        scenes_shape=scenes.shape[1:],
        conditions_shape=conditions.shape[1:],
    )
    assert disc([conditions, scenes]).shape == (5, 1)
    scenes = tf.random.normal(shape=(5, 64, 128, 1))
    conditions = tf.random.normal(shape=(5, 32, 64, 5))
    disc = dense_disc(
        scenes_shape=scenes.shape[1:],
        conditions_shape=conditions.shape[1:],
    )
    assert disc([conditions, scenes]).shape == (5, 1)


def test_unet_model():
    """ test: unet model """
    if os.environ.get("PDGM_BACK", "tf") == "pt":
        return
    scenes = tf.random.normal(shape=(5, 32, 64, 1))
    conditions = tf.random.normal(shape=(5, 32, 64, 5))
    unet_no_latent = unet_model(
        scenes_shape=scenes.shape[1:],
        conditions_shape=conditions.shape[1:],
        use_latent=False,
    )
    unet_latent = unet_model(
        scenes_shape=scenes.shape[1:],
        conditions_shape=conditions.shape[1:],
        use_latent=True,
    )
    assert unet_no_latent(conditions).shape == scenes.shape
    assert unet_latent([conditions, conditions]).shape == scenes.shape


def test_unet_shapes():
    """ test: unet model """
    if os.environ.get("PDGM_BACK", "tf") == "pt":
        return
    conditions = [
        (2**idx, 2**(jdx+1), 9) for idx, jdx in zip(range(2, 7), range(2, 7))
    ]
    scenes = [
        (2**idx, 2**(jdx+1), 3) for idx, jdx in zip(range(2, 7), range(2, 7))
    ]

    for idx, con in enumerate(conditions):
        for sce in scenes[idx:]:
            csample = tf.random.normal(shape=(5, con[0], con[1], 9))
            ssample = tf.random.normal(shape=(5, sce[0], sce[1], 3))
            unet_latent = unet_model(
                scenes_shape=sce,
                conditions_shape=con,
                use_latent=True
            )
            unet_no_latent = unet_model(
                scenes_shape=sce,
                conditions_shape=con,
                use_latent=False
            )
            assert unet_no_latent(csample).shape == ssample.shape
            assert unet_latent([csample, csample]).shape == ssample.shape



""" jouissance.tf """

from jouissance.tf.arch import set_initializer
from jouissance.tf.arch import upsample_block, downsample_block
from jouissance.tf.arch import dense_disc, unet_model
from jouissance.tf.data import MyData, make_data

__all__ = [
    "set_initializer",
    "upsample_block", "downsample_block",
    "dense_disc", "unet_model",
    "MyData", "make_data"
]
